import abc


class ModelRegistry(abc.ABC):
    @classmethod
    def __subclasshook__(cls, subclass):
        return ((hasattr(subclass, 'update_staging') and
                callable(subclass.update_staging)) and
                (hasattr(subclass, 'update_production') and
                 callable(subclass.update_production)) and
                (hasattr(subclass, 'get_model_from_registry') and
                 callable(subclass.update_production)))

    @abc.abstractmethod
    def update_staging(self, model_uri: str, model_name: str):
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def update_production(self, model_uri: str, model_name: str, model):
        """Load in the data set """
        raise NotImplementedError

    def get_model_from_registry(self, model_uri: str):
        """Load model from Registry"""
        raise NotImplementedError



import abc

import pandas as pd


class DataBaseClient(abc.ABC):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'execute_query') and
                callable(subclass.execute_query))

    @abc.abstractmethod
    def execute_query(self, query: str) -> pd.DataFrame:
        """Load in the data set"""
        raise NotImplementedError



import abc
import pandas as pd
from common.database_client import DataBaseClient
from common.file_client import FileClient
from common.logger import Logger


class Etl(abc.ABC):

    def __init__(self, data: pd.DataFrame, logger: Logger):
        self._data = data
        self._logger = logger

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'transform') and
                callable(subclass.transform))

    @classmethod
    def from_query(cls, query: str, data_base: DataBaseClient, logger: Logger):
        """Load pandas dataframe from query"""
        return cls(data_base.execute_query(query), logger)

    @classmethod
    def from_file(cls, path: str, file_client: FileClient, logger: Logger):
        """Load from file source"""
        return cls(file_client.read_from_csv(path), logger)

    @abc.abstractmethod
    def transform(self) -> pd.DataFrame:
        """Load in the data set"""
        raise NotImplementedError

    def run(self) -> pd.DataFrame:
        etl_data = self.transform()
        self._logger.log_data(self._data, 'data')
        self._logger.log_data(etl_data, 'etl_data')
        return etl_data

    def data(self) -> pd.DataFrame:
        return self._data



import abc
from common.model_registry import ModelRegistry


class Evaluation(abc.ABC):

    def __init__(self, model_registry: ModelRegistry):
        self._model_registry = model_registry

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'is_better_than_staging') and
                callable(subclass.is_better_than_staging))

    @classmethod
    def initialize(cls, model_registry: ModelRegistry):
        return cls(model_registry)

    @abc.abstractmethod
    def is_better_than_staging(self, model_uri: str, model_name: str) -> bool:
        """Load in the data set"""
        raise NotImplementedError

    def update_model(self, model_uri: str, model_name: str):
        if self.is_better_than_staging(model_uri, model_name):
            self._model_registry.update_staging(model_uri, model_name)



import abc
import pandas as pd
from typing import Optional, Dict


class Logger(abc.ABC):
    @classmethod
    def __subclasshook__(cls, subclass):
        return ((hasattr(subclass, 'log_artifact') and
                callable(subclass.log_artifact)) and
                (hasattr(subclass, 'log_data') and
                 callable(subclass.log_data)))

    @abc.abstractmethod
    def log_artifact(self, artifact: Dict, artifact_path: Optional[str] = None) -> None:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def log_data(self, data: pd.DataFrame, artifact_path: Optional[str] = None) -> None:
        """Load in the data set"""
        raise NotImplementedError



import abc
import pandas as pd
from common.logger import Logger
from typing import List


class FeatureEngineering(abc.ABC):

    def __init__(self, logger: Logger):
        self._logger = logger
        self._artifacts = []

    @classmethod
    def initialize(cls, logger: Logger):
        return cls(logger)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'transform') and
                callable(subclass.transform))

    @abc.abstractmethod
    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Load in the data set"""
        raise NotImplementedError

    def artifacts(self) -> List:
        return self._artifacts

    def run(self, data: pd.DataFrame) -> pd.DataFrame:
        transformed_data = self.transform(data)
        self.__log_artifact()
        return transformed_data

    def __log_artifact(self):
        for artifact in self._artifacts:
            self._logger.log_artifact(artifact)



import abc

from stages.etl.etl import Etl
from stages.evaluation.evaluation import Evaluation
from stages.feature_engineering.feature_engineering import FeatureEngineering
from stages.training.training import Training


class Pipeline(abc.ABC):

    def __init__(
            self,
            etl: Etl,
            feature_engineering: FeatureEngineering,
            training: Training,
            evaluation: Evaluation
    ):
        self.etl = etl
        self.feature_engineering = feature_engineering
        self.training = training
        self.evaluation = evaluation

    @classmethod
    def configure(
            cls,
            etl: Etl,
            feature_engineering: FeatureEngineering,
            training: Training,
            evaluation: Evaluation,
    ):
        return cls(etl, feature_engineering, training, evaluation)

    @abc.abstractmethod
    def _start_pipeline(self) -> None:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def _stop_pipeline(self) -> None:
        """Load in the data set"""
        raise NotImplementedError

    def run(self):
        self._start_pipeline()
        try:
            data = self.etl.run()
            data = self.feature_engineering.run(data)
            model_uri, model_name = self.training.run(data)
            self.evaluation.update_model(model_uri, model_name)
        finally:
            self._stop_pipeline()



import abc

import pandas as pd


class FileClient(abc.ABC):
    @classmethod
    def __subclasshook__(cls, subclass):
        return ((hasattr(subclass, 'read_from_csv') and
                callable(subclass.read_from_csv)) and
                (hasattr(subclass, 'write_to_csv') and
                 callable(subclass.write_to_csv)))

    @abc.abstractmethod
    def read_from_csv(self, path: str) -> pd.DataFrame:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def write_to_csv(self, data: pd.DataFrame, path: str) -> None:
        """Write data to path"""
        raise NotImplementedError



import abc
import pandas as pd
from typing import Tuple
from common.logger import Logger


class Training(abc.ABC):

    def __init__(self, label: str, logger: Logger):

        self._label = label
        self._logger = logger
        self._data = None
        self._train_data = None
        self._test_data = None
        self._model = None
        self._artifacts = []

    @classmethod
    def initialize(cls, label: str, logger: Logger):
        return cls(label, logger)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'split_data') and
                callable(subclass.split_data) and
                hasattr(subclass, 'fit') and
                callable(subclass.fit) and
                hasattr(subclass, 'model_uri') and
                callable(subclass.model_uri) and
                hasattr(subclass, 'model_name') and
                callable(subclass.model_name) and
                hasattr(subclass, '_log_model') and
                callable(subclass._log_model) and
                hasattr(subclass, '_auto_log') and
                callable(subclass._auto_log))

    def data(self) -> pd.DataFrame:
        return self._data

    def label(self) -> str:
        return self._label

    def train_data(self) -> pd.DataFrame:
        return self._train_data

    def test_data(self) -> pd.DataFrame:
        return self._test_data

    def model(self):
        return self._model

    @abc.abstractmethod
    def split_data(self, data: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def fit(self, train_data: pd.DataFrame, test_data: pd.DataFrame, label: str):
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def model_uri(self) -> str:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def model_name(self) -> str:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def _log_model(self, model) -> None:
        """Load in the data set"""
        raise NotImplementedError

    @abc.abstractmethod
    def _auto_log(self) -> None:
        """Load in the data set"""
        raise NotImplementedError

    def run(self, data: pd.DataFrame):
        self._data = data
        print('Training is starting...')
        self._train_data, self._test_data = self.split_data(data)
        self._auto_log()
        self._model = self.fit(self._train_data, self._test_data, self._label)
        self.__log_artifact()
        print('Training is done')
        return self.model_uri(), self.model_name()

    def __log_artifact(self):
        if self.train_data is not None:
            self._logger.log_data(self._train_data, 'train_data')
        if self.test_data is not None:
            self._logger.log_data(self._test_data, 'test_data')
        for artifact in self._artifacts:
            self._logger.log_artifact(artifact)
        self._log_model(self._model)
